public static final int GENE = 0;
public static final int METABOLITE = 1;
public static final int FORWARD = 2;
public static final int REVERSE = 3;
public static final int BIDIRECTIONAL = 4;
public static final int LEGEND = 5;
public static final int LEGEND_ROLLED = 6;


class ExpressionProfile
{
  float[] _profile;
  float _min_value, _max_value;

  ExpressionProfile( float[] profile )
  { 
    _min_value = Float.MAX_VALUE;
    _max_value = Float.MIN_VALUE;

    _profile = new float[profile.length];
    for ( int i = 0; i < profile.length; i++ ) 
    {
      _profile[i] = profile[i];

      if ( Float.isNaN(_profile[i]) ) continue;

      _min_value = min( _min_value, profile[i] );
      _max_value = max( _max_value, profile[i] );
    } 

    if ( (_min_value == Float.MAX_VALUE) && (_max_value == Float.MIN_VALUE) )
      _min_value = _max_value = 0.0;
  }

  float[] profile()
  { 
    return _profile; 
  }

  float minValue()
  {
    return _min_value;
  }

  float maxValue()
  {
    return _max_value;
  }

  float value( int index )
  {
    return _profile[index];
  }
}

class PathwayNode 
{
  int _type, _direction;
  String _name, _short_name;
  float[] _correlation;
  boolean[] _have_correlation;
  boolean _have_expression_profile;
  ExpressionProfile[] _expression_profile;
  float _min_over_species, _max_over_species;

  boolean _render_name;
  boolean _break_point, _branch_point, _cycle_point;
  
  String[] _paralog_names;

  int[] _x, _y;

  // metabolite
  PathwayNode( String name )
  {
    _name = name;
    _type = METABOLITE;

    _have_expression_profile = false;
    _short_name = _name;

    _correlation = new float[3];
    _have_correlation = new boolean[3];
    for ( int i = 0; i < 3; i++ ) _have_correlation[i] = false;

    _x = new int[3];
    _y = new int[3];

    _min_over_species = Float.MAX_VALUE;
    _max_over_species = Float.MIN_VALUE;

    _render_name = false;
    _break_point = _branch_point = _cycle_point = false;
    
    _paralog_names = new String[_species.length];
    for ( int i = 0; i < _species.length; i++ )
      _paralog_names[i] = new String("NONE");
  }

  // gene gets a direction
  PathwayNode( String name, int direction )
  {
    _name = name;
    _type = GENE;
    _direction = direction;

    _have_expression_profile = false;
    _short_name = _name;

    _correlation = new float[3];
    _have_correlation = new boolean[3];
    for ( int i = 0; i < 3; i++ ) _have_correlation[i] = false;

    _x = new int[3];
    _y = new int[3];

    _min_over_species = Float.MAX_VALUE;
    _max_over_species = Float.MIN_VALUE;

    _render_name = false;
    _break_point = _branch_point = _cycle_point = false;
    
    _paralog_names = new String[_species.length];
    for ( int i = 0; i < _species.length; i++ )
      _paralog_names[i] = new String("NONE");
  }

  void setXY( int x, int y, int index )
  {
    _x[index] = x;
    _y[index] = y;
  }

  int x( int index )
  {
    return _x[index];
  }

  int y( int index )
  { 
    return _y[index];
  }

  int type()
  {
    return _type;
  }

  int direction()
  { 
    return _direction;
  }

  void correlation( float corr, int index )
  { 
    _correlation[index] = corr; 
    _have_correlation[index] = true; 
  }

  float correlation( int index )
  { 
    return _correlation[index]; 
  }

  boolean haveCorrelation( int index )
  { 
    return _have_correlation[index]; 
  }

  boolean haveExpressionProfile()
  { 
    return _have_expression_profile; 
  }

  String name()
  { 
    return _name; 
  }

  void shortName( String name )
  {
    _short_name = name;
  }

  String shortName()
  {
    return _short_name;
  }

  void renderName( boolean rn )
  {
    _render_name = rn;
  }

  boolean renderName()
  {
    return _render_name;
  }

  void breakPoint( boolean bp )
  {
    _break_point = bp;
  }
  
  String paralogName(int i)
  {
    return _paralog_names[i];
  }
  
  void paralogNames( String[] n )
  {
    for ( int i = 0; i < _species.length; i++ )
      _paralog_names[i] = new String(n[i]);;
  }

  boolean breakPoint()
  {
    return _break_point;
  }

  void branchPoint( boolean bp )
  {
    _branch_point = bp;
  }

  boolean branchPoint()
  {
    return _branch_point;
  }
  
  void cyclePoint( boolean bp )
  {
    _cycle_point = bp;
  }

  boolean cyclePoint()
  {
    return _cycle_point;
  }

  ExpressionProfile expressionProfile( int species )
  {
    return _expression_profile[species];
  }

  void createExpressionProfile( float[] profile )
  {
    _expression_profile = new ExpressionProfile[_species.length];
    float[] exp_prof = new float[_timepoints.length];
    for ( int i = 0; i < _species.length; i++ )
    {
      for ( int j = 0; j < _timepoints.length; j++ )
        exp_prof[j] = profile[i*_timepoints.length + j];
      _expression_profile[i] = new ExpressionProfile(exp_prof);
    }

    // find the min and max values over the species
    for ( int i = 0; i < _species.length; i++ )
    {
      _min_over_species = min( _min_over_species, _expression_profile[i].minValue() );
      _max_over_species = max( _max_over_species, _expression_profile[i].maxValue() );
    }

    _have_expression_profile = true;
  } 

  float minValueOverSpecies()
  {
    return _min_over_species;
  }

  float maxValueOverSpecies()
  {
    return _max_over_species;
  }

  void printNode()
  {
    if ( _type == GENE ) println( "GENE: " + _direction + " " + _name );
    else println( "METABOLITE: " + _name );
  }

  void printExpressionProfile()
  {
    println( _name );
    for ( int i = 0; i < _expression_profile.length; i++ )
      println( _expression_profile[i]._profile );
  }

}

/**************************************************************/
class APathway
{
  String _name;
  boolean _have_name;
  int _start, _stop; 
  int _render_x, _render_y;
  APathway( String name, int start )
  {
    _name = name;
    _have_name = true;
    _start = start;
  }

  APathway( int start )
  {
    _have_name = false;
    _start = start;
  }

  boolean haveName()
  {
    return _have_name;
  }

  int start()
  { 
    return _start;
  }

  int stop()
  {
    return _stop;
  }

  int renderX()
  {
    return _render_x;
  }

  int renderY()
  {
    return _render_y;
  }

  String name()
  {
    return _name;
  }
}

/**************************************************************/
class Branch
{
  int _start1, _start2, _stop;
  String _branch1_name, _branch2_name, _branch1_letter, _branch2_letter;
  int _nest_level;
  int _render_y_start1, _render_y_end1, _render_y_start2, _render_y_end2;
  int _render_x1, _render_y1, _render_x2, _render_y2;
  boolean _have_name1, _have_name2;

  Branch( int start, int level )
  {
    _start1 = start;
    _nest_level = level;
    _have_name1 = _have_name2 = false;
  }

  boolean haveName1()
  {
    return _have_name1;
  }

  boolean haveName2()
  {
    return _have_name2;
  }


  int renderX1()
  {
    return _render_x1;
  }

  int renderY1()
  {
    return _render_y1;
  }

  int renderX2()
  {
    return _render_x2;
  }

  int renderY2()
  {
    return _render_y2;
  }

  int renderYStart1()
  {
    return _render_y_start1;
  }

  int renderYEnd1()
  {
    return _render_y_end1;
  }

  int renderYStart2()
  {
    return _render_y_start2;
  }

  int renderYEnd2()
  {
    return _render_y_end2;
  }

  void branchLetters( String a, String b )
  {
    _branch1_letter = a;
    _branch2_letter = b;
  }

  String branch1Letter()
  {
    return _branch1_letter;
  }

  String branch2Letter()
  {
    return _branch2_letter;
  }

  void branch1Name( String n )
  {
    _branch1_name = n;
    _have_name1 = true;
  }

  void branch2Name( String n )
  {
    _branch2_name = n;
    _have_name2 = true;
  }

  String branch1Name()
  {
    return _branch1_name;
  }

  String branch2Name()
  {
    return _branch2_name;
  }

  void start1( int s )
  {
    _start1 = s;
  }

  void start2( int s )
  {
    _start2 = s;
  }

  void stop( int s )
  {
    _stop = s;
  }

  int start1()
  {
    return _start1;
  }

  int start2()
  {
    return _start2;
  }

  int stop()
  {
    return _stop;
  }

  int nestLevel()
  {
    return _nest_level;
  }
}

/**************************************************************/
class Cycle
{
  int _start, _stop;
  String _name;
  int _render_y_start, _render_y_end;
  int _render_x, _render_y;
  boolean _have_name;

  Cycle( int start )
  {
    _start = start;
    _have_name = false;
  }

  boolean haveName()
  {
    return _have_name;
  }

  int renderX()
  {
    return _render_x;
  }

  int renderY()
  {
    return _render_y;
  }

  int renderYStart()
  {
    return _render_y_start;
  }

  int renderYEnd()
  {
    return _render_y_end;
  }

  void cycleName( String n )
  {
    _name = n;
    _have_name = true;
  }

  String cycleName()
  {
    return _name;
  }

  void start( int s )
  {
    _start = s;
  }

  void stop( int s )
  {
    _stop = s;
  }

  int start()
  {
    return _start;
  }

  int stop()
  {
    return _stop;
  }

}


/**************************************************************/
class Pathway
{
  ArrayList _nodes;
  ArrayList _pathways;
  ArrayList _branches;
  ArrayList _cycles;
  float _correlation_min, _correlation_max;
  int _num_correlation;
  int _num_genes, _num_metabolites, _num_break_points, _num_branch_points, _num_cycles;
  String[] _correlation_names;
  ArrayList[] _correlation_species;

  Pathway()
  { 
    _nodes = new ArrayList();
    _pathways = new ArrayList();
    _branches = new ArrayList();
    _cycles = new ArrayList();
    _num_genes = _num_metabolites = _num_branch_points = _num_cycles = 0;

    _correlation_min = Float.MAX_VALUE;
    _correlation_max = Float.MIN_VALUE;

    _correlation_names = new String[3];
    for ( int i = 0; i < 3; i++ ) _correlation_names[i] = "correlation" + i;
    
    _correlation_species = new ArrayList[3];
    for ( int i = 0; i < 3; i++ ) _correlation_species[i] = new ArrayList();
  }

  void correlationName( String n, int index )
  {
    _correlation_names[index] = n;
  }

  String correlationName( int index )
  {
    return _correlation_names[index];
  }

  void numCorrelation( int n )
  {
    _num_correlation = n;
  }

  int numCorrelation()
  {
    return _num_correlation;
  }

  void correlationMinMax( float cmin, float cmax )
  { 
    _correlation_min = min( _correlation_min, cmin );
    _correlation_max = max( _correlation_max, cmax );
  }

  float correlationMin()
  {
    return _correlation_min;
  }

  float correlationMax()
  {
    return _correlation_max;
  }

  void setBreakPoint( int index )
  {
    if ( ((PathwayNode)_nodes.get(index)).branchPoint() )
    {
      ((PathwayNode)_nodes.get(index)).branchPoint(false);
      --_num_branch_points;
    }
    ((PathwayNode)_nodes.get(index)).breakPoint(true);
    ++_num_break_points;
  }

  void setBranchPoint( int index )
  {
    if ( !((PathwayNode)_nodes.get(index)).breakPoint() )
    {
      ((PathwayNode)_nodes.get(index)).branchPoint(true);
      ++_num_branch_points;
    }
  }

  void setCyclePoint( int index )
  {
    ((PathwayNode)_nodes.get(index)).cyclePoint(true);
    ++_num_cycles;
  }

  void setNode( PathwayNode node )
  { 
    _nodes.add( node );

    if ( node.type() == GENE ) ++_num_genes;
    else ++_num_metabolites;
  }

  void setPathway( String name, int start )
  {
    _pathways.add( new APathway(name, start) );
  }

  void setPathway( int start )
  {
    _pathways.add( new APathway(start) );
  }

  void setPathwayStart( int start, int index )
  {
    ((APathway)_pathways.get(index))._start = start;
  }

  void setPathwayStop( int stop, int index )
  {
    ((APathway)_pathways.get(index))._stop = stop;
  }

  void setBranch( Branch b )
  {
    _branches.add( b );
  }

  Branch getBranch( int index )
  {
    return (Branch)_branches.get(index);
  } 
  
  void setCycle( Cycle b )
  {
    _cycles.add( b );
  }

  Cycle getCycle( int index )
  {
    return (Cycle)_cycles.get(index);
  } 

  PathwayNode getNode( int index )
  { 
    return (PathwayNode)_nodes.get(index); 
  }

  int numNodes()
  { 
    return _nodes.size(); 
  }

  int numGenes()
  {
    return _num_genes;
  }

  int numMetabolites()
  { 
    return _num_metabolites;
  }

  int numBreakPoints()
  {
    return _num_break_points;
  }

  int numBranchPoints()
  {
    return _num_branch_points;
  }
  
  int numCyclePoints()
  {
    return _num_cycles;
  }

  APathway getPathway( int index )
  {
    return (APathway)_pathways.get(index);
  }

  int numPathways()
  {
    return _pathways.size();
  }

  int numBranches()
  {
    return _branches.size();
  }
  
  int numCycles()
  {
    return _cycles.size();
  }

  void printPathway()
  { 
    for ( int i = 0; i < _nodes.size(); i++ ) ((PathwayNode)_nodes.get(i)).printNode(); 
    ((PathwayNode)_nodes.get(0)).printExpressionProfile();
  }
}

/**************************************************************/

class TempTableForPhyTree
{
  int _x, _y;
  int[] _child1, _child2; 
  String _name;

  TempTableForPhyTree( int x, int y ) 
  {
    _child1 = new int[2];
    _child2 = new int[2];
    position( x, y );
  }

  TempTableForPhyTree( int x, int y, String n ) 
  {
    _child1 = new int[2];
    _child2 = new int[2];
    position( x, y );
    name( n );
  }

  void position( int x, int y )
  {
    _x = x; 
    _y = y;
  }

  int x()
  {
    return _x;
  }

  int y()
  {
    return _y;
  }

  void name( String n )
  {
    _name = n;
  }

  String name()
  {
    return _name;
  }

  void childrenIndices( int i1, int j1, int i2, int j2 )
  {
    _child1[0] = i1;
    _child1[1] = j1;
    _child2[0] = i2;
    _child2[1] = j2;
  }

}









